#!/bin/bash
#  $HeadURL: https://svn.pasta.freemyip.com/main/js/tags/2/bin/jsc.sh $ $LastChangedRevision: 10935 $

#  Modules
. $(miniade) || { echo "${0##*/}: ERROR: miniade failed (hint: run 'miniade' to see error)" >&2; exit 1; }

#  Configurable stuff

#  Other globals

main()
{
    local MY_ARGS
    local JSM_SHOW_QUEUE_OUTPUT PROGNAME

    #  Defaults for options

    #  Process options
    #  Note use of '--unknown-opts' to allow unknown options to be
    #  left in $*, so that the case statement below can pass them
    #  (whatever "$@") down for processing.
    miniade_process_options --help-handler=help --unknown-opts MY_ARGS "$@" && set -- "${MY_ARGS[@]}"

    #  Process arguments
    #  (complex delegation needing PROGNAME)
    miniade_get_progname PROGNAME
    case $PROGNAME in
        jsc) [ $# -ge 1 ] || miniade_bad_usage
             COMMAND=$1
             shift
             [[ $COMMAND =~ ^(quit|suspend|resume|ping|show-queue|submit-job|chres|cancel-job|chprio-job)$ ]] || miniade_error "$COMMAND: invalid command"
             ${COMMAND//-/_} "$@" ;;
        jsq) show_queue "$@" ;;
        jss) submit_job "$@" ;;
        *)   miniade_error "unexpected progname" ;;
    esac
}

quit()
{
    local PROGNAME VERBOSELEVEL

    #  Process arguments
    [ $# = 0 ] || miniade_bad_usage

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-}
}

suspend()
{
    local PROGNAME VERBOSELEVEL

    #  Process arguments
    [ $# = 0 ] || miniade_bad_usage

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-}
}

resume()
{
    local PROGNAME VERBOSELEVEL

    #  Process arguments
    [ $# = 0 ] || miniade_bad_usage

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-}
}

ping()
{
    local PROGNAME VERBOSELEVEL

    #  Process arguments
    [ $# = 0 ] || miniade_bad_usage

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-}
}

show_queue()
{
    local SHOW_JOB_STATES VERBOSELEVEL PROGNAME

    #  Defaults for options
    SHOW_JOB_STATES=running,pending

    #  Process options
    special_opts_handler()
    {
        case $1 in
            -a) SHOW_JOB_STATES=running,pending,cancelled,completed ;;
            *)  return 1 ;;
        esac
    }
    miniade_process_options --special-opts-handler=special_opts_handler MY_ARGS "$@" && set -- "${MY_ARGS[@]}"

    #  Process arguments
    [ $# = 0 ] || miniade_bad_usage

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    miniade_debug 10 "show_queue: command=${FUNCNAME[0]//_/-} show-job-states=$SHOW_JOB_STATES"
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-} verboselevel=$VERBOSELEVEL show-job-states=$SHOW_JOB_STATES
}

cancel_job()
{
    local VERBOSELEVEL JOB_ID PROGNAME

    #  Process arguments
    [ $# = 1 ] || miniade_bad_usage
    JOB_ID=$1

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-} job-id=$JOB_ID
}

chprio_job()
{
    local VERBOSELEVEL JOB_ID PRIORITY PROGNAME

    #  Process arguments
    [ $# = 2 ] || miniade_bad_usage
    JOB_ID=$1
    PRIORITY=$2

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-} job-id=$JOB_ID priority=$PRIORITY
}

chres()
{
    local VERBOSELEVEL RESOURCE_LIST PROGNAME

    #  Process arguments
    [ $# = 1 ] || miniade_bad_usage
    RESOURCE_LIST=$1

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-} resource-list="$RESOURCE_LIST"
}

submit_job()
{
    local VERBOSELEVEL RESOURCE_LIST JOB_NAME JOB_COMMAND

    #  Defaults for options
    RESOURCE_LIST=
    JOB_NAME=unset

    #  Process options
    special_opts_handler()
    {
        case $1 in
            -r) RESOURCE_LIST=$2; extra_shift ;;
            -n) JOB_NAME=$2;      extra_shift ;;
            *)  return 1 ;;
        esac
    }
    miniade_process_options --special-opts-handler=special_opts_handler MY_ARGS "$@" && set -- "${MY_ARGS[@]}"

    #  Process arguments
    [ $# = 1 ] || miniade_bad_usage
    JOB_COMMAND="$1"

    #  Sanity checks and derivations
    miniade_get_progname PROGNAME
    miniade_get_verboselevel VERBOSELEVEL

    #  Guts
    jsm --progname=$PROGNAME --debug=$VERBOSELEVEL command=${FUNCNAME[0]//_/-} job-command="$JOB_COMMAND" resource-list="$RESOURCE_LIST" job-name="$JOB_NAME"
}

help()
{
    local PROGNAME

    miniade_get_progname PROGNAME
    help_$PROGNAME
    #  exit is done by each usage_*() function for better alignment with other scripts.
}

help_jsc()
{
    local PROGNAME

    miniade_get_progname PROGNAME
    echo "Usage:    $PROGNAME [ <standard-options> ] <cmd> ..."
    echo
    echo "Commands: quit"
    echo "          suspend"
    echo "          resume"
    echo "          ping"
    echo "          show-queue [ -a ]"
    echo "          submit-job -r <res>:<qty>[,...] <cmd>"
    echo "          cancel-job <job-id>"
    echo "          chprio-job <job-id> <priority>"
    echo "          chres <res>:<qty>[,...]"
    echo
    exit 0
}

help_jsq()
{
    local PROGNAME

    miniade_get_progname PROGNAME
    echo "Usage:    $PROGNAME [ <standard-options> ] [ -a ]"
    echo
    echo "Options:  -a             show all jobs"
    echo
    exit 0
}

help_jss()
{
    local PROGNAME

    miniade_get_progname PROGNAME
    echo "Usage:    $PROGNAME [ <standard-options> ] -r <res>:<qty>[,...] [ -n <name> ] <cmd>"
    echo
    echo "Options:  -r             specify job's resource requirements"
    echo "          -n             specify job name"
    echo
    exit 0
}

version()
{
    local PROGNAME

    miniade_get_progname PROGNAME
    echo "$PROGNAME version 0"
    exit 0
}

listpaths()
{
    exit 0
}

main "$@"
