#!/bin/bash
#  $HeadURL: https://svn.pasta.freemyip.com/main/js/tags/2/tests/bin/JS205_shepherd_child_management.sh $ $LastChangedRevision: 10273 $

#  Note this error message is sent to the terminal as it is imperative the user understands the failure.
[ $(id -u) = 0 ] || { echo "${0##*/}: this test can only pass when run as root" >/dev/tty; exit 1; }

echo "setting up environment ..."
export JS_NAME=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 8)
export JS_SHELL=$ADETEST_MODROOT/share/shepherd/js-shepherd
export JS_SHEPHERD_WARNINGS_ENABLED=false     #  make js-shepherd quieter

echo "defining support functions ..."
wait_for_jobs()
{
    while [ $(jsq | sed 1,3d | egrep -c '(running|pending)') != 0 ]; do
        sleep 0.5
    done
}

echo "starting jsd suspended ..."
#  Since the memory used by the jobs will decrease significantly below
#  the limit (because of the jss commandlines below) then it's possible
#  that all the jobs will run at the same time. That's okay, but it'll be
#  easier to debug if they run in serial. So we add an extra resource,
#  'slots', of which there is only one and which will ensure they run
#  in series.
jsd --suspend -r slots:1 2>&1 | sed -r -e 's/(child processes) \([1-9][0-9, ]*\) (are still running)/\1 (<pid>) \2/' &
#jsd --suspend -r slots:1 &
JSD_PID=$!
sleep 5  #  settle time

echo "submitting a job that spawns children ..."
#  make jss verbose because we're interested to see that the job ID it
#  mentions is the same as the one that js-shepherd complains has left
#  child processes running.
jss -v -r slots:1 "bash -c 'sleep 1234 &'"

echo "resuming jsd and waiting for jobs to complete ..."
jsc resume
wait_for_jobs

echo "cleaning up ..."
jsc quit
wait $JSD_PID

echo "all done"
