#!/bin/bash
#  $HeadURL: https://svn.pasta.freemyip.com/main/js/tags/2/tests/bin/JS206_shepherd_cpu.sh $ $LastChangedRevision: 10273 $
#  There is no point in running this test with REQUIRED_CPUS set to 1.
REQUIRED_CPUS=2

#  Note this error message is sent to the terminal as it is imperative the user understands the failure. These
#  checks are duplicated in testlist.
[ $(id -u) = 0 ] || { echo "${0##*/}: this test can only pass when run as root" >/dev/tty; exit 1; }
(( $(lscpu | sed -rn 's/^CPU\(s\): *//p') >= $REQUIRED_CPUS )) || { echo "${0##*/}: this test can only pass when run on a system with at least $REQUIRED_CPUS CPUs" >/dev/tty; exit 1; }

echo "setting up environment ..."
export JS_NAME=$(tr -dc A-Za-z0-9 </dev/urandom | head -c 8)
export JS_SHELL=$ADETEST_MODROOT/share/shepherd/js-shepherd
export JS_SHEPHERD_WARNINGS_ENABLED=false     #  make js-shepherd quieter

echo "defining support functions ..."
wait_for_jobs()
{
    while [ $(jsq | sed 1,3d | egrep -c '(running|pending)') != 0 ]; do
        sleep 0.5
    done
}

echo "starting jsd ..."
jsd --suspend -r cpu:$REQUIRED_CPUS &
JSD_PID=$!
sleep 5  #  settle time

echo "submitting $REQUIRED_CPUS jobs each using 1 CPU each ..."
FILTER1=$(printf "%05d" $RANDOM)
for X in $(seq 1 $REQUIRED_CPUS); do
    FILTER2=$(printf "%05d" $RANDOM)
    jss -r cpu:1 "export FILTER1=$FILTER1; export FILTER2=$FILTER2; for T in {1..10}; do timeout 20 bash -c 'while :; do /bin/true; done' & done; wait"
done
#jsq 

echo "letting jobs start ..."
jsc resume
sleep 5

echo "looking for processes, examining the job-specific FILTER2 value and the CPU that each is running on ..."
#  This loop is extracting info about processes that are running only for very short periods of time. So
#  we need to (1) discard error messages and (2) only display results when there were no errors.
grep -l FILTER1=$FILTER1 /proc/[0-9]*/environ 2>/dev/null | cut -d/ -f3 | while read PID; do
    FILTER2=$(strings /proc/$PID/environ 2>/dev/null | sed -n -r 's/^FILTER2=([0-9]*).*/\1/p')
    CPU=$(cut -d' ' -f39 /proc/$PID/stat 2>/dev/null)
    if [ "X$FILTER2" = X ] || [ "X$CPU" = X ]; then
        continue
    fi
    echo "$FILTER2 $CPU"
done | sort | uniq -c | sort -n | while read COUNT FILTER2 CPU; do
    echo "CPU $CPU has $COUNT pids related to one particular job on it"
done | sort
#  That should output the value of $REQUIRED_CPUS - i.e. each of the $REQUIRED_CPUS group of ~10 pids 
#  should be running on a single and different CPU.

echo "waiting for jobs to complete ..."
wait_for_jobs

echo "cleaning up ..."
jsc quit
wait $JSD_PID

echo "all done"
